from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
import requests

from pytest_httpserver.hooks import Chain
from pytest_httpserver.hooks import Delay
from pytest_httpserver.hooks import Garbage

if TYPE_CHECKING:
    from werkzeug import Request
    from werkzeug import Response

    from pytest_httpserver import HTTPServer


class MyDelay(Delay):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.evidence: int | None = None

    def _sleep(self):
        assert self.evidence is None, "_sleep should be called only once"
        self.evidence = self._seconds


def suffix_hook_factory(suffix: bytes):
    def hook(_request: Request, response: Response) -> Response:
        response.set_data(response.get_data() + suffix)
        return response

    return hook


def test_hook(httpserver: HTTPServer):
    my_hook = suffix_hook_factory(b"-SUFFIX")
    httpserver.expect_request("/foo").with_post_hook(my_hook).respond_with_data("OK")

    assert requests.get(httpserver.url_for("/foo")).text == "OK-SUFFIX"


def test_delay_hook(httpserver: HTTPServer):
    delay = MyDelay(10)
    httpserver.expect_request("/foo").with_post_hook(delay).respond_with_data("OK")
    assert requests.get(httpserver.url_for("/foo")).text == "OK"
    assert delay.evidence == 10


def test_garbage_hook(httpserver: HTTPServer):
    httpserver.expect_request("/prefix").with_post_hook(Garbage(prefix_size=128)).respond_with_data("OK")
    httpserver.expect_request("/suffix").with_post_hook(Garbage(suffix_size=128)).respond_with_data("OK")
    httpserver.expect_request("/both").with_post_hook(Garbage(prefix_size=128, suffix_size=128)).respond_with_data("OK")
    httpserver.expect_request("/large_prefix").with_post_hook(Garbage(prefix_size=10 * 1024 * 1024)).respond_with_data(
        "OK"
    )

    resp_content = requests.get(httpserver.url_for("/prefix")).content
    assert len(resp_content) == 130
    assert resp_content[128:] == b"OK"

    resp_content = requests.get(httpserver.url_for("/large_prefix")).content
    assert len(resp_content) == 10 * 1024 * 1024 + 2
    assert resp_content[10 * 1024 * 1024 :] == b"OK"

    resp_content = requests.get(httpserver.url_for("/suffix")).content
    assert len(resp_content) == 130
    assert resp_content[:2] == b"OK"

    resp_content = requests.get(httpserver.url_for("/both")).content
    assert len(resp_content) == 258
    assert resp_content[128:130] == b"OK"

    with pytest.raises(AssertionError, match="prefix_size should be positive integer"):
        Garbage(-10)

    with pytest.raises(AssertionError, match="suffix_size should be positive integer"):
        Garbage(10, -10)


def test_chain(httpserver: HTTPServer):
    delay = MyDelay(10)
    httpserver.expect_request("/foo").with_post_hook(Chain(delay, Garbage(128))).respond_with_data("OK")
    assert len(requests.get(httpserver.url_for("/foo")).content) == 130
    assert delay.evidence == 10


def test_multiple_hooks(httpserver: HTTPServer):
    delay = MyDelay(10)
    httpserver.expect_request("/foo").with_post_hook(delay).with_post_hook(Garbage(128)).respond_with_data("OK")
    assert len(requests.get(httpserver.url_for("/foo")).content) == 130
    assert delay.evidence == 10


def test_multiple_hooks_correct_order(httpserver: HTTPServer):
    hook1 = suffix_hook_factory(b"-S1")
    hook2 = suffix_hook_factory(b"-S2")

    httpserver.expect_request("/foo").with_post_hook(hook1).with_post_hook(hook2).respond_with_data("OK")

    assert requests.get(httpserver.url_for("/foo")).text == "OK-S1-S2"



import pytest
import requests

from pytest_httpserver import HTTPServer
from pytest_httpserver import RequestMatcher


def test_verify(httpserver: HTTPServer):
    httpserver.expect_request("/foo").respond_with_data("OK")
    httpserver.expect_request("/bar").respond_with_data("OKOK")

    assert list(httpserver.iter_matching_requests(httpserver.create_matcher("/foo"))) == []
    assert requests.get(httpserver.url_for("/foo")).text == "OK"
    assert requests.get(httpserver.url_for("/bar")).text == "OKOK"

    matching_log = list(httpserver.iter_matching_requests(httpserver.create_matcher("/foo")))
    assert len(matching_log) == 1

    request, response = matching_log[0]

    assert request.url == httpserver.url_for("/foo")
    assert response.get_data() == b"OK"

    assert httpserver.get_matching_requests_count(httpserver.create_matcher("/foo")) == 1
    httpserver.assert_request_made(httpserver.create_matcher("/foo"))
    httpserver.assert_request_made(httpserver.create_matcher("/no_match"), count=0)

    with pytest.raises(AssertionError):
        assert httpserver.assert_request_made(httpserver.create_matcher("/no_match"))

    with pytest.raises(AssertionError):
        assert httpserver.assert_request_made(httpserver.create_matcher("/foo"), count=2)


def test_verify_assert_msg(httpserver: HTTPServer):
    httpserver.no_handler_status_code = 404
    httpserver.expect_request("/foo", json={"foo": "bar"}, method="POST").respond_with_data("OK")
    headers = {"User-Agent": "requests", "Accept-Encoding": "gzip, deflate"}
    assert requests.get(httpserver.url_for("/foo"), headers=headers).status_code == 404

    expected_lines = [
        "Matching request found 0 times but expected 1 times.",
        "Expected request: <RequestMatcher uri='/foo' method='POST' query_string=None headers={} data"
        "=None json={'foo': 'bar'}>",
        "Found 1 similar request(s):",
        "--- Similar Request Start",
        "Path: /foo",
        "Method: GET",
        "Body: b''",
        f"Headers: Host: {httpserver.host}:{httpserver.port}",
        "User-Agent: requests",
        "Accept-Encoding: gzip, deflate",
        "Accept: */*",
        "Connection: keep-alive",
        "",
        "",
        "Query String: ''",
        "--- Similar Request End",
    ]

    with pytest.raises(AssertionError) as err:
        httpserver.assert_request_made(RequestMatcher("/foo", json={"foo": "bar"}, method="POST"))

    actual_lines = [x.strip() for x in str(err.value).splitlines()][: len(expected_lines)]
    assert actual_lines == expected_lines


def test_verify_assert_msg_no_similar_requests(httpserver: HTTPServer):
    httpserver.expect_request("/foo", json={"foo": "bar"}, method="POST").respond_with_data("OK")

    expected_lines = [
        "Matching request found 0 times but expected 1 times.",
        "Expected request: <RequestMatcher uri='/foo' method='POST' query_string=None headers={} data"
        "=None json={'foo': 'bar'}>",
        "No similar requests found.",
    ]

    with pytest.raises(AssertionError) as err:
        httpserver.assert_request_made(RequestMatcher("/foo", json={"foo": "bar"}, method="POST"))

    actual_lines = [x.strip() for x in str(err.value).splitlines()][: len(expected_lines)]
    assert actual_lines == expected_lines



import requests

from pytest_httpserver import HTTPServer
from pytest_httpserver.hooks import Delay


def test_delay(httpserver: HTTPServer):
    # this adds 0.5 seconds delay to the server response
    httpserver.expect_request("/foo").with_post_hook(Delay(0.5)).respond_with_json({"example": "foo"})

    assert requests.get(httpserver.url_for("/foo")).json() == {"example": "foo"}



import requests
from werkzeug import Request
from werkzeug import Response

from pytest_httpserver import HTTPServer


def my_hook(_request: Request, response: Response) -> Response:
    # add a new header value to the response
    response.headers["X-Example"] = "Example"
    return response


def test_custom_hook(httpserver: HTTPServer):
    httpserver.expect_request("/foo").with_post_hook(my_hook).respond_with_data(b"OK")

    assert requests.get(httpserver.url_for("/foo")).headers["X-Example"] == "Example"



import http.client
import threading
import time
from collections.abc import Iterable

import pytest
from werkzeug import Request
from werkzeug import Response

from pytest_httpserver import HTTPServer


@pytest.fixture
def threaded() -> Iterable[HTTPServer]:
    server = HTTPServer(threaded=True)
    server.start()
    yield server
    server.clear()
    if server.is_running():
        server.stop()


def test_threaded(threaded: HTTPServer):
    sleep_time = 0.5

    def handler(_request: Request):
        # allow some time to the client to have multiple pending request
        # handlers running in parallel
        time.sleep(sleep_time)

        # send back thread id
        return Response(f"{threading.get_ident()}")

    threaded.expect_request("/foo").respond_with_handler(handler)

    number_of_connections = 5
    conns = [http.client.HTTPConnection(threaded.host, threaded.port) for _ in range(number_of_connections)]

    for conn in conns:
        conn.request("GET", "/foo", headers={"Host": threaded.host})

    thread_ids: list[int] = []
    for conn in conns:
        response = conn.getresponse()

        assert response.status == 200
        thread_ids.append(int(response.read()))

    for conn in conns:
        conn.close()

    assert len(thread_ids) == len(set(thread_ids)), "thread ids returned should be unique"



import requests

from pytest_httpserver import HTTPServer
from pytest_httpserver import RequestMatcher


def test_log_querying_example(httpserver: HTTPServer):
    # set up the handler for the request
    httpserver.expect_request("/foo").respond_with_data("OK")

    # make a request matching the handler
    assert requests.get(httpserver.url_for("/foo")).text == "OK", "Response should be 'OK'"

    # make another request non-matching and handler
    assert (
        requests.get(httpserver.url_for("/no_match")).status_code == 500
    ), "Response code should be 500 for non-matched requests"

    # you can query the log directly
    # log will contain all request-response pair, including non-matching
    # requests and their response as well
    assert len(httpserver.log) == 2, "2 request-response pairs should be in the log"

    # there are the following methods to query the log
    #
    # each one uses the matcher we created for the handler in the very beginning
    # of this test, RequestMatcher accepts the same parameters what you were
    # specifying to the `expect_request` (and similar) methods.

    # 1. get counts
    # (returns 0 for non-matches)
    httpserver.get_matching_requests_count(
        RequestMatcher("/foo")
    ) == 1, "There should be one request matching the the /foo request"

    # 2. assert for matching request counts
    # by default it asserts for exactly 1 matches
    # it is roughly the same as:
    # ```
    # assert httpserver.get_matching_requests_count(...) == 1
    # ```
    # assertion text will be a fully-detailed explanation about the error, including
    # the similar handlers (which might have been inproperly configured)
    httpserver.assert_request_made(RequestMatcher("/foo"))

    # you can also specify the counts
    # if you want, you can specify 0 to check for non-matching requests

    # there should have been 0 requests for /bar
    httpserver.assert_request_made(RequestMatcher("/bar"), count=0)

    # 3. iterate over the matching request-response pairs
    # this provides you greater flexibility
    for request, response in httpserver.iter_matching_requests(RequestMatcher("/foo")):
        assert request.url == httpserver.url_for("/foo")
        assert response.get_data() == b"OK"



# this is required to make sphinx able to find references for classes put inside
# typing.TYPE_CHECKING block

from ssl import SSLContext

from werkzeug import Request
from werkzeug import Response

import pytest_httpserver.blocking_httpserver
import pytest_httpserver.httpserver

pytest_httpserver.httpserver.SSLContext = SSLContext
pytest_httpserver.blocking_httpserver.SSLContext = SSLContext

pytest_httpserver.blocking_httpserver.Request = Request
pytest_httpserver.blocking_httpserver.Response = Response



import pytest
import requests

from pytest_httpserver import HTTPServer


class Client:
    def __init__(self) -> None:
        self.url: str | None = None

    def get(self):
        if self.url:
            requests.get(self.url)


@pytest.fixture
def my_fixture():
    client = Client()
    yield client
    client.get()


def test_1(my_fixture: Client, httpserver: HTTPServer):
    httpserver.expect_request("/foo").respond_with_data("OK")
    my_fixture.url = httpserver.url_for("/foo")


def test_2(httpserver: HTTPServer):
    assert httpserver.log == []



import requests
from werkzeug import Request

from pytest_httpserver import HTTPServer
from pytest_httpserver import RequestMatcher


class MyMatcher(RequestMatcher):
    def match(self, request: Request) -> bool:
        match = super().match(request)
        if not match:  # existing parameters didn't match -> return with False
            return match

        # match the json's "value" key: if it is an integer and it is an even
        # number, it returns True
        json = request.json
        if isinstance(json, dict) and isinstance(json.get("value"), int):
            return json["value"] % 2 == 0

        return False


def test_custom_request_matcher(httpserver: HTTPServer):
    httpserver.expect(MyMatcher("/foo")).respond_with_data("OK")

    # with even number it matches the request
    resp = requests.post(httpserver.url_for("/foo"), json={"value": 42})
    resp.raise_for_status()
    assert resp.text == "OK"

    resp = requests.post(httpserver.url_for("/foo"), json={"value": 198})
    resp.raise_for_status()
    assert resp.text == "OK"

    # with an odd number, it does not match the request
    resp = requests.post(httpserver.url_for("/foo"), json={"value": 43})
    assert resp.status_code == 500



import requests

from pytest_httpserver import HTTPServer


def test_expect_method(httpserver: HTTPServer):
    expected_response = "OK"
    matcher = httpserver.create_matcher(uri="/test", method="POST")
    httpserver.expect(matcher).respond_with_data(expected_response)
    resp = requests.post(httpserver.url_for("/test"), json={"list": [1, 2, 3, 4]})
    assert resp.text == expected_response



from __future__ import annotations

import email
import re
import shutil
import subprocess
import tarfile
import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from collections.abc import Iterable

try:
    import tomllib
except ImportError:
    # Unfortunately mypy cannot handle this try/expect pattern, and "type: ignore"
    # is the simplest work-around. See: https://github.com/python/mypy/issues/1153
    import tomli as tomllib  # type: ignore

# TODO: skip if poetry is not available or add mark to test it explicitly


pytestmark = pytest.mark.release

NAME = "pytest-httpserver"
NAME_UNDERSCORE = NAME.replace("-", "_")
PY_MAX_VERSION = (3, 13)


@pytest.fixture(scope="session")
def pyproject_path() -> Path:
    return Path("pyproject.toml")


@pytest.fixture(scope="session")
def pyproject(pyproject_path: Path):
    assert pyproject_path.is_file()
    with pyproject_path.open("rb") as infile:
        pyproject = tomllib.load(infile)
    return pyproject


class Wheel:
    def __init__(self, path: Path):
        self.path = path

    @property
    def wheel_out_dir(self) -> Path:
        return self.path.parent.joinpath("wheel")

    def extract(self):
        with zipfile.ZipFile(self.path) as zf:
            zf.extractall(self.wheel_out_dir)  # noqa: S202

    def get_meta(self, version: str, name: str = NAME_UNDERSCORE) -> email.message.Message:
        metadata_path = self.wheel_out_dir.joinpath(f"{name}-{version}.dist-info", "METADATA")
        with metadata_path.open() as metadata_file:
            msg = email.message_from_file(metadata_file)

        return msg


class Sdist:
    def __init__(self, path: Path):
        self.path = path

    @property
    def sdist_out_dir(self) -> Path:
        return self.path.parent.joinpath("sdist")

    def extract(self):
        with tarfile.open(self.path, mode="r:gz") as tf:
            tf.extractall(self.sdist_out_dir)  # noqa: S202


@dataclass
class Build:
    wheel: Wheel
    sdist: Sdist

    def extract(self):
        self.wheel.extract()
        self.sdist.extract()


@pytest.fixture(scope="session")
def build() -> Iterable[Build]:
    dist_path = Path("dist").resolve()
    if dist_path.is_dir():
        shutil.rmtree(dist_path)

    try:
        subprocess.run(["poetry", "build"], check=True)
        assert dist_path.is_dir()
        wheels = list(dist_path.glob("*.whl"))
        sdists = list(dist_path.glob("*.tar.gz"))
        assert len(wheels) == 1
        assert len(sdists) == 1
        build = Build(wheel=Wheel(wheels[0]), sdist=Sdist(sdists[0]))
        build.extract()
        yield build

    finally:
        shutil.rmtree(dist_path)


@pytest.fixture(scope="session")
def version(pyproject) -> str:
    return pyproject["tool"]["poetry"]["version"]


def version_to_tuple(version: str) -> tuple:
    return tuple([int(x) for x in version.split(".")])


def test_no_duplicate_classifiers(build: Build, pyproject):
    pyproject_meta = pyproject["tool"]["poetry"]
    wheel_meta = build.wheel.get_meta(version=pyproject_meta["version"])
    classifiers = wheel_meta.get_all("Classifier")
    assert classifiers is not None

    sorted_classifiers = sorted(classifiers)
    unique_classifiers = sorted(set(classifiers))

    assert sorted_classifiers == unique_classifiers


def test_python_version(build: Build, pyproject):
    pyproject_meta = pyproject["tool"]["poetry"]
    wheel_meta = build.wheel.get_meta(version=pyproject_meta["version"])
    python_dependency = pyproject_meta["dependencies"]["python"]
    m = re.match(r">=(\d+\.\d+)", python_dependency)
    if m:
        min_version, *_ = m.groups()
    else:
        raise ValueError(python_dependency)

    min_version_tuple = version_to_tuple(min_version)

    classifiers = wheel_meta.get_all("Classifier")
    assert classifiers is not None

    for classifier in classifiers:
        if classifier.startswith("Programming Language :: Python ::"):
            version_tuple = version_to_tuple(classifier.split("::")[-1].strip())
            if len(version_tuple) > 1:
                assert version_tuple >= min_version_tuple
                assert version_tuple <= PY_MAX_VERSION


def test_wheel_no_extra_contents(build: Build, version: str):
    wheel_dir = build.wheel.wheel_out_dir
    wheel_contents = list(wheel_dir.iterdir())
    assert len(wheel_contents) == 2
    assert wheel_dir.joinpath(NAME_UNDERSCORE).is_dir()
    assert wheel_dir.joinpath(f"{NAME_UNDERSCORE}-{version}.dist-info").is_dir()

    package_contents = {path.name for path in wheel_dir.joinpath(NAME_UNDERSCORE).iterdir()}
    assert package_contents == {
        "__init__.py",
        "blocking_httpserver.py",
        "hooks.py",
        "httpserver.py",
        "py.typed",
        "pytest_plugin.py",
    }


def test_sdist_contents(build: Build, version: str):
    sdist_base = build.sdist.sdist_out_dir.joinpath(f"pytest_httpserver-{version}")

    subdir_contents = {
        ".": {
            "CHANGES.rst",
            "CONTRIBUTION.md",
            "doc",
            "example_pytest.py",
            "example.py",
            "LICENSE",
            "PKG-INFO",
            "pyproject.toml",
            "pytest_httpserver",
            "README.md",
            "tests",
        },
        "doc": {
            "_static",
            "api.rst",
            "background.rst",
            "changes.rst",
            "conf.py",
            "fixtures.rst",
            "guide.rst",
            "howto.rst",
            "index.rst",
            "Makefile",
            "patch.py",
            "tutorial.rst",
            "upgrade.rst",
        },
        "pytest_httpserver": {
            "__init__.py",
            "blocking_httpserver.py",
            "hooks.py",
            "httpserver.py",
            "py.typed",
            "pytest_plugin.py",
        },
        "tests": {
            "assets",
            "conftest.py",
            "examples",
            "test_blocking_httpserver.py",
            "test_handler_errors.py",
            "test_headers.py",
            "test_hooks.py",
            "test_ip_protocols.py",
            "test_json_matcher.py",
            "test_log_leak.py",
            "test_log_querying.py",
            "test_mixed.py",
            "test_oneshot.py",
            "test_ordered.py",
            "test_permanent.py",
            "test_parse_qs.py",
            "test_port_changing.py",
            "test_querymatcher.py",
            "test_querystring.py",
            "test_release.py",
            "test_ssl.py",
            "test_thread_type.py",
            "test_threaded.py",
            "test_urimatch.py",
            "test_wait.py",
            "test_with_statement.py",
            "test_matcher.py",
        },
    }

    for subdir, subdir_content in subdir_contents.items():
        contents = {path.name for path in sdist_base.joinpath(subdir).iterdir()}
        assert contents == subdir_content


def test_poetry_check():
    subprocess.run(["poetry", "check"], check=True)



#!/usr/bin/env python3

import argparse
import subprocess
import sys
from collections.abc import Iterable
from pathlib import Path
from shutil import which


class UsageError(Exception):
    pass


def parse_version() -> str:
    output = subprocess.check_output(["poetry", "version", "--short"], encoding="utf-8")
    version = output.strip()

    return version


def bump_version(path: Path, prefix_list: Iterable[str], current_version: str, new_version: str):
    prefixes = tuple(prefix_list)
    lines = []
    for line in path.open():
        if line.startswith(prefixes):
            line = line.replace(current_version, new_version)
        lines.append(line)

    path.write_text("".join(lines))


def git(*args):
    return subprocess.check_call(["git"] + list(args))


def make(*args):
    return subprocess.check_call(["make"] + list(args))


def check_changelog():
    old_changelog = Path("CHANGES.rst").read_text()
    make("changes")
    new_changelog = Path("CHANGES.rst").read_text()
    if old_changelog == new_changelog:
        raise UsageError("No new changelog entries")


def check_environment():
    for binary in ("git", "make", "poetry"):
        if not which(binary):
            raise UsageError("No such binary: {}".format(binary))


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("new_version", help="Version to release")

    args = parser.parse_args()
    new_version: str = args.new_version

    current_version = parse_version()
    if current_version is None:
        raise UsageError("Unable to parse version")

    print(f"Current version: {current_version}")

    if current_version == new_version:
        raise UsageError("Current version is the same as new version")

    check_changelog()

    bump_version(Path("doc/conf.py"), ["version"], current_version, new_version)
    subprocess.check_call(["poetry", "version", new_version])

    git("add", "pyproject.toml", "doc/conf.py")
    git("commit", "-m", "Version bump to {}".format(new_version))
    git("tag", new_version)
    make("changes")
    git("add", "CHANGES.rst")
    git("commit", "-m", "CHANGES.rst: add release notes for {}".format(new_version))
    git("tag", "-f", new_version)


if __name__ == "__main__":
    try:
        main()
    except UsageError as err:
        print(f"ERROR: {err}", file=sys.stderr)
        sys.exit(1)



from __future__ import annotations

import threading
from typing import TYPE_CHECKING

import requests
from werkzeug import Response

if TYPE_CHECKING:
    from werkzeug import Request

    from pytest_httpserver import HTTPServer


def test_server_thread_is_daemon(httpserver: HTTPServer):
    def handler(_request: Request):
        return Response(f"{threading.current_thread().daemon}")

    httpserver.expect_request("/foo").respond_with_handler(handler)

    assert requests.get(httpserver.url_for("/foo")).text == "True"
